#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# -------------------------------------------------------------------------------
#                                                                               -
#  Python dual-logging setup (console and log file),                            -
#  supporting different log levels and colorized output                         -
#                                                                               -
#  Created by Fonic <https://github.com/fonic>                                  -
#  Date: 04/05/20 - 02/07/23                                                    -
#                                                                               -
#  Based on:                                                                    -
#  https://stackoverflow.com/a/13733863/1976617                                 -
#  https://uran198.github.io/en/python/2016/07/12/colorful-python-logging.html  -
#  https://en.wikipedia.org/wiki/ANSI_escape_code#Colors                        -
#                                                                               -
# -------------------------------------------------------------------------------
# Imports
import logging
import os
import sys
import ctypes
from logging.handlers import RotatingFileHandler

# Enable ANSI terminal mode for Command Prompt on Microsoft Windows


def windows_enable_ansi_terminal_mode():
    if (sys.platform != "win32"):
        return None
    try:
        kernel32 = ctypes.windll.kernel32
        result = kernel32.SetConsoleMode(kernel32.GetStdHandle(-11), 7)
        if (result == 0):
            raise Exception
        return True
    except Exception as e:
        logging.exception(f'Error when enableing terminal mode: {e}')
        return False

# Imports

# Logging formatter supporting colorized output


class LogFormatter(logging.Formatter):

    COLOR_CODES = {
        logging.CRITICAL: "\033[1;35m",  # bright/bold magenta
        logging.ERROR: "\033[1;31m",  # bright/bold red
        logging.WARNING: "\033[1;33m",  # bright/bold yellow
        logging.INFO: "\033[0;37m",  # white / light gray
        logging.DEBUG: "\033[1;30m"  # bright/bold black / dark gray
    }

    RESET_CODE = "\033[0m"

    def __init__(self, color, *args, **kwargs):
        super(LogFormatter, self).__init__(*args, **kwargs)
        self.color = color

    def format(self, record, *args, **kwargs):
        if (self.color is True and record.levelno in self.COLOR_CODES):
            record.color_on = self.COLOR_CODES[record.levelno]
            record.color_off = self.RESET_CODE
        else:
            record.color_on = ""
            record.color_off = ""
        return super(LogFormatter, self).format(record, *args, **kwargs)

# Set up logging


def set_up_logging(console_log_output, console_log_level, console_log_color, logfile_file, logfile_log_level, logfile_log_color, warn_file, warn_log_level, warn_log_color, log_line_template):

    # Create logger
    # For simplicity, we use the root logger, i.e. call 'logging.getLogger()'
    # without name argument. This way we can simply use module methods for
    # for logging throughout the script. An alternative would be exporting
    # the logger, i.e. 'global logger; logger = logging.getLogger("<name>")'
    logger = logging.getLogger()

    # Set global log level to 'debug' (required for handler levels to work)
    logger.setLevel(logging.DEBUG)

    # Create console handler
    console_log_output = console_log_output.lower()
    if (console_log_output == "stdout"):
        console_log_output = sys.stdout
    elif (console_log_output == "stderr"):
        console_log_output = sys.stderr
    else:
        logging.error("Failed to set console output: invalid output: '%s'" % console_log_output)
        return False
    console_handler = logging.StreamHandler(console_log_output)

    # Set console log level
    try:
        console_handler.setLevel(console_log_level.upper())  # only accepts uppercase level names
    except Exception:
        logging.exception("Failed to set console log level: invalid level: '%s'" % console_log_level)
        return False

    # Create and set formatter, add console handler to logger
    console_formatter = LogFormatter(fmt=log_line_template, color=console_log_color)
    console_handler.setFormatter(console_formatter)
    logger.addHandler(console_handler)

    # Create log file handler
    maxBytes = 20 * 1024 * 1024
    backupCount = 20

    try:
        logfile_handler = RotatingFileHandler(logfile_file, maxBytes=maxBytes, backupCount=backupCount, encoding='utf-8')
    except Exception as exception:
        logging.exception("Failed to set up log file: %s" % str(exception))
        return False

    # Set log file log level
    try:
        logfile_handler.setLevel(logfile_log_level.upper())  # only accepts uppercase level names
    except Exception:
        logging.exception("Failed to set log file log level: invalid level: '%s'" % logfile_log_level)
        return False

    # Create and set formatter, add log file handler to logger
    logfile_formatter = LogFormatter(fmt=log_line_template, color=logfile_log_color)
    logfile_handler.setFormatter(logfile_formatter)
    logger.addHandler(logfile_handler)

    # Create log file handler
    try:
        warn_handler = RotatingFileHandler(warn_file, maxBytes=maxBytes, backupCount=backupCount, encoding='utf-8')
    except Exception as exception:
        logging.exception("Failed to set up log file: %s" % str(exception))
        return False

    # Set log file log level
    try:
        warn_handler.setLevel(warn_log_level.upper())  # only accepts uppercase level names
    except Exception:
        logging.exception("Failed to set log file log level: invalid level: '%s'" % warn_log_level)
        return False

    # Create and set formatter, add log file handler to logger
    warn_formatter = LogFormatter(fmt=log_line_template, color=warn_log_color)
    warn_handler.setFormatter(warn_formatter)
    logger.addHandler(warn_handler)

    # Success
    return True

# Main function


def main():

    # Set up logging
    script_name = os.path.splitext(os.path.basename(sys.argv[0]))[0]
    if (not set_up_logging(console_log_output="stdout", console_log_level="info", console_log_color=True,
                           logfile_file=script_name + ".log", logfile_log_level="debug", logfile_log_color=False,
                           warn_file=script_name + "warn.log", warn_log_level="warning", warn_log_color=False,
                           log_line_template="%(color_on)s[%(asctime)s] [%(threadName)s] [%(levelname)-8s] %(message)s%(color_off)s")):
        logging.error("Failed to set up logging, aborting.")
        return 1

    # Log some messages
    logging.debug("Debug message")
    logging.info("Info message")
    logging.warning("Warning message")
    logging.error("Error message")
    logging.critical("Critical message")


# Call main function
if (__name__ == "__main__"):
    sys.exit(main())



import pytest
from unittest.mock import patch
import pandas as pd
from exifScan.CreateMetadata import MetadataWriter
from datetime import datetime as dt, timedelta as td
import os
import logging


@pytest.fixture
def setup_writer():
    WorkDirectory = 'tests/test_folders/metadata'
    metashapeCacheTime = 10
    MetashapeDirectory = '/path/to/metashape'
    os.makedirs(WorkDirectory, exist_ok=True)

    def removeGdf():
        pass

    writer = MetadataWriter(WorkDirectory, metashapeCacheTime, MetashapeDirectory, removeGdf)
    gpdDf = pd.DataFrame({
        'SourceFileDir': ['/path/to/source1', '/path/to/source2']
    })
    return writer, gpdDf


@patch('exifScan.CreateMetadata.os.listdir')
@patch('exifScan.CreateMetadata.pd.read_excel')
def test_locate_and_match_metadata(mock_read_excel, mock_listdir, setup_writer):
    writer, gpdDf = setup_writer
    mock_listdir.return_value = ['metadata.xlsx', '']
    mock_read_excel.return_value = pd.DataFrame({
        'Index': ['col1', 'col2'],
        'Values': ['val1', 'val2']
    }).set_index('Index')

    result = writer.locate_and_match_metadata(gpdDf.copy())
    assert 'Metadata File Location' in result.columns
    print(result)
    assert result.loc[result['SourceFileDir'] == '/path/to/source1', 'Metadata File Location'].values[0] == '\\path\\to\\source1'


@patch('exifScan.CreateMetadata.PsxLocator.main')
@patch('exifScan.CreateMetadata.pickle.load')
@patch('exifScan.CreateMetadata.pickle.dump')
@patch('exifScan.CreateMetadata.os.path.exists')
@patch('exifScan.CreateMetadata.os.path.getmtime')
def test_associate_metashape_projects(mock_getmtime, mock_exists, mock_dump, mock_load, mock_psxlocator, setup_writer, caplog):
    writer, gpdDf = setup_writer
    mock_exists.return_value = False
    mock_psxlocator.return_value = {'/path/to/source1': 'project1'}
    mock_load.return_value = {'/path/to/source1': 'project1'}
    mock_getmtime.return_value = (dt.now() - td(minutes=5)).timestamp()

    with caplog.at_level(logging.INFO):
        result = writer.associate_metashape_projects(gpdDf.copy())
        assert 'Metashape Files' in result.columns
        assert result.loc[result['SourceFileDir'] == '/path/to/source1', 'Metashape Files'].values[0] == 'project1'
        assert 'Using cached Metashape project matches' in caplog.text or 'Metashape data was saved to' in caplog.text or '1 Metashape project file(s) contain images from the database.' in caplog.text



import pytest
import pandas as pd
import numpy as np
from exifScan.geoDfFromExifDf import GeoDataProcessor
import logging


@pytest.fixture
def setup_processor():
    WorkDirectory = '/path/to/workdir'
    xyTagNames = ['GPSLongitude', 'GPSLatitude']
    zTagName = 'GPSAltitude'
    groupByList = ['CameraModel', 'LensModel']
    processor = GeoDataProcessor(WorkDirectory, xyTagNames, zTagName, groupByList)
    exifDf = pd.DataFrame({
        'SourceFile': ['/path/to/source1/file1.jpg', '/path/to/source2/file2.jpg'],
        'GPSLongitude': [174.7633, np.nan],
        'GPSLatitude': [-36.8485, np.nan],
        'GPSAltitude': [10, np.nan],
        'CreateDate': ['2021:01:01', '2021:01:02'],
        'CameraModel': ['Model1', 'Model2'],
        'LensModel': ['Lens1', 'Lens2']
    })
    return processor, exifDf


def test_process_valid_data(setup_processor, caplog):
    processor, exifDf = setup_processor

    with caplog.at_level(logging.INFO):
        pointsDf, polygonsDf, nonGeoDf = processor.process(exifDf.copy())

        assert not pointsDf.empty
        assert not polygonsDf.empty
        assert 'Handling geolocated images' in caplog.text or 'Points created' in caplog.text or 'Grouped points into polygons' in caplog


def test_process_invalid_data(setup_processor, caplog):
    processor, exifDf = setup_processor
    exifDf['GPSLongitude'] = np.nan
    exifDf['GPSLatitude'] = np.nan

    with caplog.at_level(logging.INFO):
        pointsDf, polygonsDf, nonGeoDf = processor.process(exifDf.copy())

        assert pointsDf.empty
        assert polygonsDf.empty
        assert not nonGeoDf.empty
        assert 'Handling non-geolocated images' in caplog.text or 'images are not geolocated' in caplog.text


def test_process_mixed_data(setup_processor, caplog):
    processor, exifDf = setup_processor
    exifDf.loc[1, 'GPSLongitude'] = 174.7633
    exifDf.loc[1, 'GPSLatitude'] = -36.8485

    with caplog.at_level(logging.INFO):
        pointsDf, polygonsDf, nonGeoDf = processor.process(exifDf.copy())

        assert not pointsDf.empty
        assert not polygonsDf.empty
        assert nonGeoDf.empty
        assert 'Handling geolocated images' in caplog.text or 'images are geolocated' in caplog.text



import pytest
from unittest.mock import patch, MagicMock
import os
import glob
from exifScan.runExifScan import ExifScanner
import tempfile
import shutil
import geopandas as gpd
import pandas as pd
import pyogrio
from shapely.geometry import Point, Polygon
gpd.options.io_engine = "pyogrio"


@pytest.fixture
def setup_test_environment():
    # Create a temporary directory
    test_dir = tempfile.mkdtemp()
    yield test_dir
    # Cleanup after tests
    shutil.rmtree(test_dir)


def test_initialization(setup_test_environment):
    test_dir = setup_test_environment
    scanner = ExifScanner(WorkDirectory=test_dir)

    assert scanner.WorkDirectory == test_dir
    assert scanner.OutputGeopackage == 'Photos.gpkg'
    assert scanner.imageFileExtensions == ['JPG']


def test_scan_for_photos_no_folder(setup_test_environment):
    test_dir = setup_test_environment
    scanner = ExifScanner(WorkDirectory=test_dir)

    with pytest.raises(Exception) as excinfo:
        scanner.scan_for_photos(None)
    assert "Error: ScanFolder is not defined or is False" in str(excinfo.value)


def test_scan_for_photos_empty_folder(setup_test_environment):
    test_dir = setup_test_environment
    empty_folder = os.path.join(test_dir, 'empty_folder')
    os.makedirs(empty_folder)

    scanner = ExifScanner(WorkDirectory=test_dir)
    result = scanner.scan_for_photos(empty_folder)

    assert result == empty_folder


@patch('subprocess.run')
def test_update_data_kart_not_initialized(mock_subprocess_run, setup_test_environment):
    test_dir = setup_test_environment

    # Mock subprocess.run to simulate Kart not initialized
    mock_subprocess_run.side_effect = [
        MagicMock(returncode=0),   # call to exiftool check
        MagicMock(returncode=41),  # First call to kart import
        MagicMock(returncode=0)    # Second call to kart init
    ]
    scanner = ExifScanner(WorkDirectory=test_dir, KartDirectory=test_dir)

    scanner.kart_writer.import_to_kart()
    scanner.write_HTML()

    # Check that subprocess.run was called with the expected commands
    mock_subprocess_run.assert_any_call(
        ["kart", "import", scanner.OutputGeopackagePath, "--all-tables", "--replace-existing"],
        cwd=test_dir
    )
    mock_subprocess_run.assert_any_call(["exiftool", "-ver"])
    mock_subprocess_run.assert_any_call(['kart', '--version'], stdout=-1, stderr=-1)
    mock_subprocess_run.assert_any_call(
        ['kart', 'import', scanner.OutputGeopackagePath, '--all-tables', '--replace-existing'],
        cwd=test_dir
    )


@patch('subprocess.run')
def test_update_data_kart_import_success(mock_subprocess_run, setup_test_environment):
    test_dir = setup_test_environment
    mock_subprocess_run.side_effect = [
        MagicMock(returncode=0),   # call to exiftool check
        MagicMock(returncode=0),    # Second call to kart check
        MagicMock(returncode=0)    # third call to kart init
    ]
    # Mock subprocess.run to simulate successful Kart import
    scanner = ExifScanner(WorkDirectory=test_dir, KartDirectory=test_dir)

    scanner.kart_writer.import_to_kart()
    scanner.write_HTML()
    # Check that subprocess.run was called with the expected command
    mock_subprocess_run.assert_called_with(
        ["kart", "import", scanner.OutputGeopackagePath, "--all-tables", "--replace-existing"],
        cwd=test_dir
    )


@patch('subprocess.run')
def test_update_data_kart_remote_not_set(mock_subprocess_run, setup_test_environment):
    test_dir = setup_test_environment

    # Mock subprocess.run to simulate Kart remote not set
    mock_subprocess_run.side_effect = [
        MagicMock(returncode=0),  # ExifTool check
        MagicMock(returncode=0),  # kart check
        MagicMock(returncode=0),    # First call to kart import
        MagicMock(returncode=128),  # Second call to kart push
        MagicMock(returncode=0),    # Third call to kart remote add
        MagicMock(returncode=0)     # Fourth call to kart push --set-upstream
    ]
    scanner = ExifScanner(WorkDirectory=test_dir, KartDirectory=test_dir, KartRemote='remote_repo')

    scanner.kart_writer.import_to_kart()
    scanner.write_HTML()
    # Check that subprocess.run was called with the expected commands
    mock_subprocess_run.assert_any_call(
        ["kart", "import", scanner.OutputGeopackagePath, "--all-tables", "--replace-existing"],
        cwd=test_dir
    )
    mock_subprocess_run.assert_any_call(["exiftool", "-ver"])
    mock_subprocess_run.assert_any_call(['kart', '--version'], stdout=-1, stderr=-1)
    mock_subprocess_run.assert_any_call(
        ['kart', 'push'],
        cwd=test_dir
    )
    mock_subprocess_run.assert_any_call(
        ['kart', 'remote', 'add', 'origin', 'remote_repo'],
        cwd=test_dir
    )
    mock_subprocess_run.assert_any_call(
        ['kart', 'push', '--set-upstream', 'origin', 'main'],
        cwd=test_dir
    )


@patch('geopandas.read_file')
@patch('pyogrio.list_layers')
@patch('geopandas.GeoDataFrame.to_file')
def test_removefromGdf(mock_to_file, mock_list_layers, mock_read_file, setup_test_environment):
    test_dir = setup_test_environment
    scanner = ExifScanner(WorkDirectory=test_dir)

    # Create a mock GeoDataFrame for existing data
    existing_data = {
        'SourceFileDir': ['dir1', 'dir2', 'dir3'],
        'Make': ['Canon', 'Nikon', 'Sony'],
        'Model': ['EOS', 'D850', 'A7']
    }

    # Add a geometry column with Point geometries
    geometry = [Point(1, 2), Point(2, 3), Point(3, 4)]
    existing_data['geometry'] = geometry
    existing_gdf = gpd.GeoDataFrame(existing_data, index=[1, 2, 3])
    mock_read_file.return_value = existing_gdf

    # Mock the list of layers
    mock_list_layers.return_value = [('layer1', 'type')]

    # Call the removefromGdf method
    result = scanner.removefromGdf('dir2')

    # Check that read_file was called with the correct arguments
    mock_read_file.assert_called_with(scanner.OutputGeopackagePath, layer='layer1', fid_as_index=True)
    mock_to_file.assert_called_with(scanner.OutputGeopackagePath, layer='layer1', driver="GPKG")
    # Check that the method returns True
    assert result is True


@patch('geopandas.read_file')
def test_removefromGdf_file_not_exist(mock_read_file, setup_test_environment):
    test_dir = setup_test_environment
    scanner = ExifScanner(WorkDirectory=test_dir)

    # Mock read_file to raise DataSourceError
    mock_read_file.side_effect = pyogrio.errors.DataSourceError("File not found")

    # Call the removefromGdf method
    result = scanner.removefromGdf('dir2')

    # Check that the method returns True
    assert result is True


def test_scan_for_photos_with_images(setup_test_environment):
    test_dir = setup_test_environment
    images_folder = os.path.join(test_dir, 'images')
    os.makedirs(images_folder)

    # Copy test images to the temporary directory
    src_images_folder = 'tests/images'
    for file_name in os.listdir(src_images_folder):
        full_file_name = os.path.join(src_images_folder, file_name)
        if os.path.isfile(full_file_name):
            shutil.copy(full_file_name, images_folder)

    scanner = ExifScanner(WorkDirectory=test_dir, htmlFolder=test_dir)
    result = scanner.scan_for_photos(images_folder)
    # Check if the output files are created
    exif_output_pattern = os.path.join(test_dir, scanner.ExifOutputFile + '*.json')
    database_output_pattern = os.path.join(test_dir, 'Database*.gpkg')
    matching_exif_files = glob.glob(exif_output_pattern)
    matching_database_files = glob.glob(database_output_pattern)
    assert len(matching_exif_files) == 1  # Ensure exactly one matching file exists
    assert len(matching_database_files) == 1  # Ensure exactly one matching file exists
    assert result == images_folder
    assert os.path.exists(scanner.htmlWriter.mapLocation)
    assert os.path.exists(scanner.metadata_writer.metashape_cache_file)
    assert os.path.exists(scanner.OutputGeopackagePath)
    output = gpd.read_file(scanner.OutputGeopackagePath)
    # Define the 3D polygon coordinates
    coords1 = [
        (172.676957944444, -43.6096788333333, -201.204),
        (172.677330055556, -43.6089029444444, -204.459),
        (172.677897861111, -43.6084271944444, -209.174),
        (172.677254277778, -43.60940975, -201.059),
        (172.676957944444, -43.6096788333333, -201.204)
    ]
    coords2 = [
        (172.759785861111, -43.5664325833333, 6.14),
        (172.759792388889, -43.5664220555556, 4.94),
        (172.759900777778, -43.5664180833333, 8.34),
        (172.759785861111, -43.5664325833333, 6.14)
    ]

    # Create the 3D polygon
    polygon1 = Polygon(coords1)
    polygon2 = Polygon(coords2)
    data = {
        'Make': ['DJI', 'DJI'],
        'Model': ['FC6310', 'FC6310'],
        'SourceFileDir': [images_folder.replace("\\", "/"), images_folder.replace("\\", "/")],
        'Image Count': [3, 4],
        'CreateDate': ['2019-02-21', '2024-07-09'],
        'areaSqkm': [9.53602682174712e-06, 0.003264685668783136],
        'geometry': [polygon2, polygon1]}
    
    decimal_places=6
    for key in output.keys():
        if key == 'areaSqkm': #rounding error on linux
            rounded_output = [round(num, decimal_places) for num in output[key]]
            rounded_data = [round(num, decimal_places) for num in data[key]]
            assert rounded_output == rounded_data, f"Values differ: {rounded_output} != {rounded_data}"
        else:
            assert output[key].tolist() == data[key],f"Values differ: {output[key].tolist()} != {data[key]}"
    # repeat to check if duplication is avoided
    result = scanner.scan_for_photos(images_folder)
    output = gpd.read_file(scanner.OutputGeopackagePath)
    for key in output.keys():
        if key == 'areaSqkm': #rounding error on linux
            rounded_output = [round(num, decimal_places) for num in output[key]]
            rounded_data = [round(num, decimal_places) for num in data[key]]
            assert rounded_output == rounded_data, f"Values differ: {rounded_output} != {rounded_data}"
        else:
            assert output[key].tolist() == data[key],f"Values differ: {output[key].tolist()} != {data[key]}"

    new_row = {
        'Make': 'Canon', 'Model': 'EOS 5D Mark IV', 'SourceFileDir': 'non_existant_folder', 'Image Count': 1, 'CreateDate': '2024-08-01', 'areaSqkm': 0.001, 'geometry': polygon1
    }
    new_data = pd.DataFrame([new_row])
    new_gdf = gpd.GeoDataFrame(new_data, geometry='geometry')
    new_gdf.to_file(scanner.OutputGeopackagePath, layer=scanner.gpkg_layer, driver='GPKG', mode='a')

    assert any(new_gdf['Make'] == 'Canon')
    assert any(new_gdf['Model'] == 'EOS 5D Mark IV')
    assert any(new_gdf['CreateDate'] == '2024-08-01')

    # Run the scan again
    result = scanner.scan_for_photos(images_folder)
    output = gpd.read_file(scanner.OutputGeopackagePath)

    # Check if the new row is still present after the scan
    assert any(output['Make'] != 'Canon')
    assert any(output['Model'] != 'EOS 5D Mark IV')
    assert any(output['CreateDate'] != '2024-08-01')

    result = scanner.removefromGdf(images_folder)

    # Check that to_file was called with the filtered GeoDataFrame
    filtered_data = {
        'Make': [], 'Model': [], 'SourceFileDir': [], 'Image Count': [], 'CreateDate': [], 'areaSqkm': [], 'geometry': []
    }
    filtered_output = gpd.read_file(scanner.OutputGeopackagePath)
    for key in filtered_output.keys():
        assert filtered_output[key].tolist() == filtered_data[key]



import subprocess
import pandas as pd
import logging
import threading
import os
import sys


class ExifToolScanner:
    def __init__(self, FileTypes, recursive=False, ignoreDirs=[], ignoreMetashape=False):
        self.FileTypes = FileTypes
        self.recursive = recursive
        self.ignoreDirs = ignoreDirs
        self.ignoreMetashape = ignoreMetashape

        check_exiftool()

    def run_exiftool(self, ScanFolder, saveFile="exifOut.json"):
        logging.info(f"Exiftool now scanning for geolocated {' '.join(self.FileTypes)} files in: {ScanFolder}")

        exifCmd = [
            "exiftool",
            "-GPSLongitude", "-GPSLatitude", "-GPSAltitude",
            "-SourceFile", "-CreateDate", "-ImageSize", "-FileSize", "-Make", "-Model",
            "-j", "-c", "%.6f", "-n",
            ScanFolder,
            "-progress"
        ]
        for filetype in self.FileTypes:
            exifCmd.append('-ext')
            exifCmd.append(filetype)

        for directory in self.ignoreDirs:
            exifCmd.append('-i')
            directory = directory.replace('\\', '/')
            exifCmd.append(directory)
        if self.ignoreMetashape:
            exifCmd.append('-if')
            exifCmd.append('$directory !~ /\\.files/')

        if self.recursive:
            exifCmd.append('-r')

        logging.info(f"Command used for Exiftool: {' '.join(exifCmd)}")
        logging.debug(f"File used for Exiftool: {saveFile}")

        with open(saveFile, "w") as outfile:
            stop_event = threading.Event()
            process = subprocess.Popen(exifCmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)

            stdout_thread = threading.Thread(target=reader, args=[process.stdout, outfile.write, stop_event])
            stderr_thread = threading.Thread(target=reader, args=[process.stderr, lambda x: logging.info(x.strip()), stop_event])

            stdout_thread.start()
            stderr_thread.start()
            try:
                process.wait()

                stdout_thread.join()
                stderr_thread.join()

            except KeyboardInterrupt:
                stop_event.set()
                process.terminate()
                stdout_thread.join()
                stderr_thread.join()
                return None

        return process_output(ScanFolder, saveFile)


def process_output(ScanFolder, saveFile):
    try:
        if os.path.getsize(saveFile) != 0:
            logging.debug(f'Reading {saveFile}')
            exif_df = pd.read_json(saveFile)
        else:
            logging.info(f"No images were found in {ScanFolder}")
            os.remove(saveFile)
            exif_df = pd.DataFrame()

    except FileNotFoundError:
        logging.error(f"FileNotFoundError: The file {saveFile} does not exist or the path is incorrect.")
        return None
    except ValueError:
        logging.info(f"The file {saveFile} might be empty (no images were found) or not a valid JSON file.")
        return None
    except Exception as e:
        logging.error(f"An unexpected error occurred when attempting to read {saveFile}: {e}")
        return None

    return exif_df


def reader(pipe, func, stop_event):
    try:
        with pipe:
            for line in iter(pipe.readline, ''):
                if stop_event.is_set():
                    break
                func(line)
    except Exception as e:
        logging.error(f'Error in reader: {str(e)}')


def check_exiftool():

    try:
        result = subprocess.run(["exiftool", "-ver"])

        if result.returncode == 0:
            logging.debug("Exiftool installed")
            return
        else:
            logging.error(f'Exiftool check return code: {result.returncode}')
    except Exception as e:
        logging.error(f'Error when checking for exiftool: {e}')

    logging.error("Error when trying to run exiftool from command line. Please check if exiftool is installed and available by running 'exiftool' from any terminal.")
    sys.exit()



from watchdog.events import FileSystemEventHandler
from watchdog.utils.dirsnapshot import EmptyDirectorySnapshot

from threading import Timer

import logging

import os

from datetime import datetime as dt


class FolderUpdateHandler(FileSystemEventHandler):
    def __init__(self, ignoreDirs, Scanner, timer_interval=30):
        self.directoriesToCheck = []
        self.paused = False
        self.timer = None
        self.ignoreDirs = ignoreDirs
        if len(self.ignoreDirs):
            logging.info(f'Watchdog is ignoring directories that contain: {self.ignoreDirs}')
        # only makes sense to read from json file at startup
        self.Scanner = Scanner
        self.timer_interval = timer_interval
        self.currentSnapshot = EmptyDirectorySnapshot()

    def appendDirectory(self, directory):
        if not directory:
            return

        if directory not in self.directoriesToCheck:
            self.directoriesToCheck.append(directory)
            logging.warning(f'Appended {directory} to directories requiring update.')

        self.restartTimer()

    def restartTimer(self):
        if self.timer:
            self.timer.cancel()
            logging.debug('Restarting timer.')
        self.timer = Timer(self.timer_interval, self.process_event)
        self.timer.start()

    def update_snapshot(self, snapshot):
        self.recentSnapshot = snapshot
        if not self.paused:
            logging.debug('currentSnapshot updated')

            self.currentSnapshot = snapshot

    def get_snapshot(self):
        return self.currentSnapshot

    def process_event(self):
        self.timer = None
        if (self.paused):
            self.restartTimer()
            logging.debug('Handler is paused.')
            return
        self.paused = True
        logging.debug('Pausing Handler.')

        logging.warning(f"Processing all change events. Directories to check: {self.directoriesToCheck}")

        directoriesToCheckCopy = self.directoriesToCheck.copy()
        for path in directoriesToCheckCopy:

            result = self.Scanner.scan_for_photos(path)
            if result != path:
                logging.error(f"An error occurred when scanning the following folder:{path}")
                logging.error("Retrying in 1 hour")
                # set limit for re-tries?
                self.directoriesToCheck.remove(path)
                Timer(3600, self.appendDirectory, [path])
            else:
                self.directoriesToCheck.remove(path)
                logging.info(f'{dt.now()} Successfully scanned the following path for images: {path}')
                logging.info('======================================================================')
        self.currentSnapshot = self.recentSnapshot
        logging.debug('Resuming Handler after updating currentSnapshot')
        logging.warning("Done processing all change events.")
        self.paused = False

    def getDirectory(self, event, path):
        if not event.is_directory:
            directory = os.path.dirname(path)
        elif event.is_directory:
            directory = path
        else:
            return
        for ignore in self.ignoreDirs:
            if ignore in directory:
                return
        return directory.rstrip('/\\')

    def on_any_event(self, event):

        directory = self.getDirectory(event, event.src_path)
        logging.info(f'Event type: {event.event_type}, event source : {event.src_path}')
        self.appendDirectory(directory)

        if (directory):
            if hasattr(event, 'dest_path'):
                directory = self.getDirectory(event, event.dest_path)
                logging.info(f'Event type: {event.event_type}, event source : {event.dest_path}')
                self.appendDirectory(directory)



import time
import os
from datetime import datetime as dt
import logging
import sys
import copy
from exifScan.runExifScan import ExifScanner
from persistentObserver.persistentObserver import PersistentPollingObserver
from persistentObserver.folderUpdateHandler import FolderUpdateHandler
import threading

startCtime = dt.now().ctime()
stats = {'Scan Started at': startCtime}


def watchFolder(path, argsDict, stop_event=threading.Event()):
    # check for a config file - will only set if not the default value. Unfortunately- this overwrites the command line if it is set to default

    pollingInterval = argsDict.pop('pollingInterval', None)
    persistObserver = argsDict.pop('persistObserver', None)
    timerInterval = argsDict.pop('timerInterval', None)
    on_snapshot = argsDict.pop('on_snapshot', None)

    ExifStartTime = dt.now()
    # deep copy the args so we can set differing args for the initial scan
    initialArgs = copy.deepcopy(argsDict)
    # pop so it isn't an argument in any of the future scans
    ignoreMetashape = argsDict.pop('ignoreMetashape', False)
    ignoreDirs = argsDict.get('ignoreDirs', [])
    if ignoreMetashape:
        ignoreDirs.append('.files')
        ignoreDirs.append('auto-project')

    # pop the exifOutputFiles from the argsDict,
    # as they are only useful in the initial scan
    readFromFile = argsDict.pop('readExifFromOutputFile', None)
    argsDict.pop('ExifOutputFile', None)
    argsDict['stats'] = stats
    logging.warning(f'Watchdog starting new process. Watching: {path}')
    # set handler and start polling.
    Scanner = ExifScanner(**argsDict)
    event_handler = FolderUpdateHandler(ignoreDirs, Scanner, timer_interval=timerInterval)
    pollingKeywords = {}
    if pollingInterval:
        pollingKeywords['timeout'] = pollingInterval

    if persistObserver:
        pollingKeywords['save_to'] = os.path.join(argsDict['WorkDirectory'], 'PersistentObserver','image.pkl')

    def on_snapshot_callback(time):
        now = dt.now().ctime()
        logging.info(f"Snapshot took {time}(h:mm:ss).")
        if not event_handler.paused:
            stats.pop('Currently updating. Update started', None)
            stats['Last updated'] = now
            Scanner.write_HTML()
        else:
            stats.pop('Last updated', None)
            stats['Currently updating. Update started'] = now
            Scanner.write_HTML()

    if on_snapshot:
        pollingKeywords['on_snapshot'] = on_snapshot_callback

    observer = PersistentPollingObserver(**pollingKeywords)

    # warning that might save a lot of time.
    gpdir = argsDict['GeopackageDirectory'] or argsDict['WorkDirectory']

    OutputGeopackagePath = os.path.join(gpdir, argsDict.get('OutputGeopackage', None))
    if not observer.previous_snapshot_exists or readFromFile:
        if not readFromFile:
            logging.warning('No previous snapshot, will create a full initial recursive scan with Exiftool')
        else:
            logging.warning('Will use previous ExifScan output file.')

        if os.path.exists(OutputGeopackagePath):
            logging.error(f'{OutputGeopackagePath} already exists, and a full scan has been initiated. Please confirm this is intentional.')
            while True:
                confirmation = input("Are you sure you want to proceed? (y/n): ")
                if confirmation.lower() == 'y':
                    break
                elif confirmation.lower() == 'n':
                    logging.warning("Aborting script.")
                    sys.exit()
                else:
                    logging.error("Invalid input. Please enter 'y' or 'n'.")
    elif observer.previous_snapshot_exists:
        logging.warning('Re-using previous snapshot.')
        if readFromFile:
            logging.error('Ignoring previous ExifFile, previous snapshot file exists!')

    observer.schedule(event_handler, path=path, recursive=True)
    logging.debug('Starting Handler.')

    # check if previous snapshot exists, if not scan recursively.
    if not observer.previous_snapshot_exists:
        observer.start()
        logging.info(f"First Folder Poll done . The first poll took {dt.now() - ExifStartTime}(h:mm:ss).")
        initialArgs['recursive'] = True
        initialScanner = ExifScanner(**initialArgs)
        result = initialScanner.scan_for_photos(path)

        if result != path:
            logging.error(f'Error while performing initial scan in folder {path}')
        else:
            logging.info(f"Successfully scanned the following path for images: {path}")
            logging.info('======================================================================')
            try:
                while not stop_event.is_set():
                    time.sleep(1)
            except KeyboardInterrupt:
                logging.info("KeyboardInterrupt")
                observer.stop()
                observer.join()

    else:
        observer.start()
        logging.info(f"First Folder Poll done . The first poll took {dt.now() - ExifStartTime}(h:mm:ss).")

        try:
            while not stop_event.is_set():
                time.sleep(1)
        except KeyboardInterrupt:
            logging.info("KeyboardInterrupt")
            observer.stop()
            observer.join()

    logging.info('Script ended')
